@@ -52,6 +52,7 @@ from version import *
 from keystore import load_keystore, Hardware_KeyStore
 from storage import multisig_type
 
+import transaction
 from transaction import Transaction
 from plugins import run_hook
 import bitcoin
@@ -61,8 +62,9 @@ from verifier import SPV
 from mnemonic import Mnemonic
 
 import paymentrequest
+from paymentrequest import InvoiceStore
+from contacts import Contacts
 
-from storage import WalletStorage
 
 TX_STATUS = [
     _('Replaceable'),
@@ -127,6 +129,11 @@ class Abstract_Wallet(PrintError):
         if self.storage.get('wallet_type') is None:
             self.storage.put('wallet_type', self.wallet_type)
 
+        # invoices and contacts
+        self.invoices = InvoiceStore(self.storage)
+        self.contacts = Contacts(self.storage)
+
+
     def diagnostic_name(self):
         return self.basename()
 
@@ -208,15 +215,13 @@ class Abstract_Wallet(PrintError):
     def basename(self):
         return os.path.basename(self.storage.path)
 
-    def save_pubkeys(self):
-        self.storage.put('pubkeys', {'receiving':self.receiving_pubkeys, 'change':self.change_pubkeys})
+    def save_addresses(self):
+        self.storage.put('addresses', {'receiving':self.receiving_addresses, 'change':self.change_addresses})
 
     def load_addresses(self):
-        d = self.storage.get('pubkeys', {})
-        self.receiving_pubkeys = d.get('receiving', [])
-        self.change_pubkeys = d.get('change', [])
-        self.receiving_addresses = map(self.pubkeys_to_address, self.receiving_pubkeys)
-        self.change_addresses = map(self.pubkeys_to_address, self.change_pubkeys)
+        d = self.storage.get('addresses', {})
+        self.receiving_addresses = d.get('receiving', [])
+        self.change_addresses = d.get('change', [])
 
     def synchronize(self):
         pass
@@ -234,6 +239,7 @@ class Abstract_Wallet(PrintError):
         changed = False
         old_text = self.labels.get(name)
         if text:
+            text = text.replace("\n", " ")
             if old_text != text:
                 self.labels[name] = text
                 changed = True
@@ -258,24 +264,15 @@ class Abstract_Wallet(PrintError):
 
     def get_address_index(self, address):
         if self.keystore.can_import():
-            i = self.receiving_addresses.index(address)
-            return self.receiving_pubkeys[i]
+            for pubkey in self.keystore.keypairs.keys():
+                if self.pubkeys_to_address(pubkey) == address:
+                    return pubkey
         elif address in self.receiving_addresses:
             return False, self.receiving_addresses.index(address)
         if address in self.change_addresses:
             return True, self.change_addresses.index(address)
         raise Exception("Address not found", address)
 
-    def get_pubkey_index(self, pubkey):
-        if self.keystore.can_import():
-            assert pubkey in self.receiving_pubkeys
-            return pubkey
-        elif pubkey in self.receiving_pubkeys:
-            return False, self.receiving_pubkeys.index(pubkey)
-        if pubkey in self.change_pubkeys:
-            return True, self.change_pubkeys.index(pubkey)
-        raise Exception("Pubkey not found", pubkey)
-
     def get_private_key(self, address, password):
         if self.is_watching_only():
             return []
@@ -285,8 +282,7 @@ class Abstract_Wallet(PrintError):
 
     def get_public_key(self, address):
         if self.keystore.can_import():
-            i = self.receiving_addresses.index(address)
-            pubkey = self.receiving_pubkeys[i]
+            pubkey = self.get_address_index(address)
         else:
             sequence = self.get_address_index(address)
             pubkey = self.get_pubkey(*sequence)
@@ -435,8 +431,8 @@ class Abstract_Wallet(PrintError):
         can_bump = False
         label = ''
         height = conf = timestamp = None
+        tx_hash = tx.txid()
         if tx.is_complete():
-            tx_hash = tx.hash()
             if tx_hash in self.transactions.keys():
                 label = self.get_label(tx_hash)
                 height, conf, timestamp = self.get_tx_height(tx_hash)
@@ -460,7 +456,6 @@ class Abstract_Wallet(PrintError):
         else:
             s, r = tx.signature_count()
             status = _("Unsigned") if s == 0 else _('Partially signed') + ' (%d/%d)'%(s,r)
-            tx_hash = None
 
         if is_relevant:
             if is_mine:
@@ -618,7 +613,7 @@ class Abstract_Wallet(PrintError):
                 if _type == TYPE_ADDRESS:
                     addr = x
                 elif _type == TYPE_PUBKEY:
-                    addr = public_key_to_bc_address(x.decode('hex'))
+                    addr = bitcoin.public_key_to_p2pkh(x.decode('hex'))
                 else:
                     addr = None
                 if addr and self.is_mine(addr):
@@ -794,10 +789,6 @@ class Abstract_Wallet(PrintError):
         # Change <= dust threshold is added to the tx fee
         return 182 * 3 * self.relayfee() / 1000
 
-    def get_tx_fee(self, tx):
-        # this method can be overloaded
-        return tx.get_fee()
-
     def make_unsigned_transaction(self, inputs, outputs, config, fixed_fee=None, change_addr=None):
         # check outputs
         i_max = None
@@ -815,6 +806,9 @@ class Abstract_Wallet(PrintError):
         if not inputs:
             raise NotEnoughFunds()
 
+        if fixed_fee is None and config.fee_per_kb() is None:
+            raise BaseException('Dynamic fee estimates not available')
+
         for item in inputs:
             self.add_input_info(item)
 
@@ -879,10 +873,11 @@ class Abstract_Wallet(PrintError):
             pubkey = public_key_from_private_key(privkey)
             address = address_from_private_key(privkey)
             u = network.synchronous_get(('blockchain.address.listunspent', [address]))
-            pay_script = Transaction.pay_script(TYPE_ADDRESS, address)
+            pay_script = transaction.get_scriptPubKey(address)
             for item in u:
                 if len(inputs) >= imax:
                     break
+                item['type'] = 'p2pkh'
                 item['scriptPubKey'] = pay_script
                 item['redeemPubkey'] = pubkey
                 item['address'] = address
@@ -1042,7 +1037,27 @@ class Abstract_Wallet(PrintError):
             raise BaseException(_('Cannot bump fee: cound not find suitable outputs'))
         return Transaction.from_io(inputs, outputs)
 
+    def cpfp(self, tx, fee):
+        txid = tx.txid()
+        for i, o in enumerate(tx.outputs()):
+            otype, address, value = o
+            if otype == TYPE_ADDRESS and self.is_mine(address):
+                break
+        else:
+            return
+        coins = self.get_addr_utxo(address)
+        for item in coins:
+            if item['prevout_hash'] == txid and item['prevout_n'] == i:
+                break
+        else:
+            return
+        self.add_input_info(item)
+        inputs = [item]
+        outputs = [(TYPE_ADDRESS, address, value - fee)]
+        return Transaction.from_io(inputs, outputs)
+
     def add_input_info(self, txin):
+        txin['type'] = self.txin_type
         # Add address for utxo that are in wallet
         if txin.get('scriptSig') == '':
             coins = self.get_spendable_coins()
@@ -1117,6 +1132,8 @@ class Abstract_Wallet(PrintError):
     def get_receiving_address(self):
         # always return an address
         domain = self.get_receiving_addresses()
+        if not domain:
+            return
         choice = domain[0]
         for addr in domain:
             if not self.history.get(addr):
@@ -1302,6 +1319,7 @@ class Imported_Wallet(Abstract_Wallet):
     # wallet made of imported addresses
 
     wallet_type = 'imported'
+    txin_type = 'unknown'
 
     def __init__(self, storage):
         Abstract_Wallet.__init__(self, storage)
@@ -1377,49 +1395,11 @@ class Imported_Wallet(Abstract_Wallet):
 
     def add_input_sig_info(self, txin, address):
         addrtype, hash160 = bc_address_to_hash_160(address)
-        xpubkey = 'fd' + (chr(addrtype) + hash160).encode('hex')
-        txin['x_pubkeys'] = [ xpubkey ]
-        txin['pubkeys'] = [ xpubkey ]
-        txin['signatures'] = [None]
-
-
-class P2PKH_Wallet(Abstract_Wallet):
-
-    def pubkeys_to_address(self, pubkey):
-        return public_key_to_bc_address(pubkey.decode('hex'))
-
-    def load_keystore(self):
-        self.keystore = load_keystore(self.storage, 'keystore')
-
-    def get_pubkey(self, c, i):
-        pubkey_list = self.change_pubkeys if c else self.receiving_pubkeys
-        return pubkey_list[i]
-
-    def get_public_keys(self, address):
-        return [self.get_public_key(address)]
-
-    def add_input_sig_info(self, txin, address):
-        if not self.keystore.can_import():
-            txin['derivation'] = derivation = self.get_address_index(address)
-            x_pubkey = self.keystore.get_xpubkey(*derivation)
-            pubkey = self.get_pubkey(*derivation)
-        else:
-            pubkey = self.get_public_key(address)
-            assert pubkey is not None
-            x_pubkey = pubkey
+        x_pubkey = 'fd' + (chr(addrtype) + hash160).encode('hex')
         txin['x_pubkeys'] = [x_pubkey]
-        txin['pubkeys'] = [pubkey]
         txin['signatures'] = [None]
-        txin['redeemPubkey'] = pubkey
-        txin['num_sig'] = 1
 
-    def sign_message(self, address, message, password):
-        index = self.get_address_index(address)
-        return self.keystore.sign_message(index, message, password)
 
-    def decrypt_message(self, pubkey, message, password):
-        index = self.get_pubkey_index(pubkey)
-        return self.keystore.decrypt_message(index, message, password)
 
 
 class Deterministic_Wallet(Abstract_Wallet):
@@ -1456,11 +1436,10 @@ class Deterministic_Wallet(Abstract_Wallet):
             addresses = self.get_receiving_addresses()
             k = self.num_unused_trailing_addresses(addresses)
             n = len(addresses) - k + value
-            self.receiving_pubkeys = self.receiving_pubkeys[0:n]
             self.receiving_addresses = self.receiving_addresses[0:n]
             self.gap_limit = value
             self.storage.put('gap_limit', self.gap_limit)
-            self.save_pubkeys()
+            self.save_addresses()
             return True
         else:
             return False
@@ -1476,7 +1455,7 @@ class Deterministic_Wallet(Abstract_Wallet):
         # fixme: this assumes wallet is synchronized
         n = 0
         nmax = 0
-        addresses = self.account.get_receiving_addresses()
+        addresses = self.get_receiving_addresses()
         k = self.num_unused_trailing_addresses(addresses)
         for a in addresses[0:-k]:
             if self.history.get(a):
@@ -1487,14 +1466,12 @@ class Deterministic_Wallet(Abstract_Wallet):
         return nmax + 1
 
     def create_new_address(self, for_change):
-        pubkey_list = self.change_pubkeys if for_change else self.receiving_pubkeys
-        n = len(pubkey_list)
-        x = self.new_pubkeys(for_change, n)
-        pubkey_list.append(x)
-        self.save_pubkeys()
-        address = self.pubkeys_to_address(x)
         addr_list = self.change_addresses if for_change else self.receiving_addresses
+        n = len(addr_list)
+        x = self.derive_pubkeys(for_change, n)
+        address = self.pubkeys_to_address(x)
         addr_list.append(address)
+        self.save_addresses()
         self.add_address(address)
         return address
 
@@ -1516,10 +1493,10 @@ class Deterministic_Wallet(Abstract_Wallet):
                 self.synchronize_sequence(False)
                 self.synchronize_sequence(True)
             else:
-                if len(self.receiving_pubkeys) != len(self.keystore.keypairs):
-                    self.receiving_pubkeys = self.keystore.keypairs.keys()
-                    self.save_pubkeys()
-                    self.receiving_addresses = map(self.pubkeys_to_address, self.receiving_pubkeys)
+                if len(self.receiving_addresses) != len(self.keystore.keypairs):
+                    pubkeys = self.keystore.keypairs.keys()
+                    self.receiving_addresses = map(self.pubkeys_to_address, pubkeys)
+                    self.save_addresses()
                     for addr in self.receiving_addresses:
                         self.add_address(addr)
 
@@ -1545,8 +1522,42 @@ class Deterministic_Wallet(Abstract_Wallet):
 
 
 
-class Standard_Wallet(Deterministic_Wallet, P2PKH_Wallet):
-    wallet_type = 'standard'
+class Simple_Wallet(Abstract_Wallet):
+
+    """ Wallet with a single pubkey per address """
+
+    def load_keystore(self):
+        self.keystore = load_keystore(self.storage, 'keystore')
+        self.is_segwit = self.keystore.is_segwit()
+        self.txin_type = 'p2wpkh-p2sh' if self.is_segwit else 'p2pkh'
+
+    def get_pubkey(self, c, i):
+        return self.derive_pubkeys(c, i)
+
+    def get_public_keys(self, address):
+        return [self.get_public_key(address)]
+
+    def add_input_sig_info(self, txin, address):
+        if not self.keystore.can_import():
+            derivation = self.get_address_index(address)
+            x_pubkey = self.keystore.get_xpubkey(*derivation)
+        else:
+            x_pubkey = self.get_public_key(address)
+        txin['x_pubkeys'] = [x_pubkey]
+        txin['signatures'] = [None]
+        txin['num_sig'] = 1
+
+    def sign_message(self, address, message, password):
+        index = self.get_address_index(address)
+        return self.keystore.sign_message(index, message, password)
+
+    def decrypt_message(self, pubkey, message, password):
+        addr = self.pubkeys_to_address(pubkey)
+        index = self.get_address_index(addr)
+        return self.keystore.decrypt_message(index, message, password)
+
+
+class Simple_Deterministic_Wallet(Deterministic_Wallet, Simple_Wallet):
 
     def __init__(self, storage):
         Deterministic_Wallet.__init__(self, storage)
@@ -1554,7 +1565,7 @@ class Standard_Wallet(Deterministic_Wallet, P2PKH_Wallet):
     def get_master_public_key(self):
         return self.keystore.get_master_public_key()
 
-    def new_pubkeys(self, c, i):
+    def derive_pubkeys(self, c, i):
         return self.keystore.derive_pubkey(c, i)
 
     def get_keystore(self):
@@ -1572,10 +1583,10 @@ class Standard_Wallet(Deterministic_Wallet, P2PKH_Wallet):
     def check_password(self, password):
         self.keystore.check_password(password)
 
-    def update_password(self, old_pw, new_pw):
+    def update_password(self, old_pw, new_pw, encrypt=False):
         self.keystore.update_password(old_pw, new_pw)
         self.save_keystore()
-        self.storage.put('use_encryption', (new_pw is not None))
+        self.storage.set_password(new_pw, encrypt)
         self.storage.write()
 
     def save_keystore(self):
@@ -1588,8 +1599,8 @@ class Standard_Wallet(Deterministic_Wallet, P2PKH_Wallet):
         pubkey = self.get_public_key(address)
         self.keystore.delete_imported_key(pubkey)
         self.save_keystore()
-        self.receiving_pubkeys.remove(pubkey)
         self.receiving_addresses.remove(address)
+        self.save_addresses()
         self.storage.write()
 
     def can_import_privkey(self):
@@ -1598,18 +1609,48 @@ class Standard_Wallet(Deterministic_Wallet, P2PKH_Wallet):
     def import_key(self, pk, pw):
         pubkey = self.keystore.import_key(pk, pw)
         self.save_keystore()
-        self.receiving_pubkeys.append(pubkey)
-        self.save_pubkeys()
         addr = self.pubkeys_to_address(pubkey)
         self.receiving_addresses.append(addr)
+        self.save_addresses()
         self.storage.write()
         self.add_address(addr)
         return addr
 
 
-class Multisig_Wallet(Deterministic_Wallet):
+class P2SH:
+
+    def pubkeys_to_redeem_script(self, pubkeys):
+        raise NotImplementedError()
+
+    def pubkeys_to_address(self, pubkey):
+        redeem_script = self.pubkeys_to_redeem_script(pubkey)
+        return bitcoin.hash160_to_p2sh(hash_160(redeem_script.decode('hex')))
+
+
+class Standard_Wallet(Simple_Deterministic_Wallet):
+    wallet_type = 'standard'
+
+    def pubkeys_to_redeem_script(self, pubkey):
+        if self.is_segwit:
+            return transaction.segwit_script(pubkey)
+
+    def pubkeys_to_address(self, pubkey):
+        if not self.is_segwit:
+            return bitcoin.public_key_to_p2pkh(pubkey.decode('hex'))
+        elif bitcoin.TESTNET:
+            redeem_script = self.pubkeys_to_redeem_script(pubkey)
+            return bitcoin.hash160_to_p2sh(hash_160(redeem_script.decode('hex')))
+        else:
+            raise NotImplementedError()
+
+
+
+
+
+class Multisig_Wallet(Deterministic_Wallet, P2SH):
     # generic m of n
     gap_limit = 20
+    txin_type = 'p2sh'
 
     def __init__(self, storage):
         self.wallet_type = storage.get('wallet_type')
@@ -1617,19 +1658,16 @@ class Multisig_Wallet(Deterministic_Wallet):
         Deterministic_Wallet.__init__(self, storage)
 
     def get_pubkeys(self, c, i):
-        pubkey_list = self.change_pubkeys if c else self.receiving_pubkeys
-        return pubkey_list[i]
+        return self.derive_pubkeys(c, i)
 
     def redeem_script(self, c, i):
         pubkeys = self.get_pubkeys(c, i)
-        return Transaction.multisig_script(sorted(pubkeys), self.m)
+        return transaction.multisig_script(sorted(pubkeys), self.m)
 
-    def pubkeys_to_address(self, pubkeys):
-        redeem_script = Transaction.multisig_script(sorted(pubkeys), self.m)
-        address = hash_160_to_bc_address(hash_160(redeem_script.decode('hex')), bitcoin.ADDRTYPE_P2SH)
-        return address
+    def pubkeys_to_redeem_script(self, pubkeys):
+        return transaction.multisig_script(sorted(pubkeys), self.m)
 
-    def new_pubkeys(self, c, i):
+    def derive_pubkeys(self, c, i):
         return [k.derive_pubkey(c, i) for k in self.get_keystores()]
 
     def load_keystore(self):
@@ -1649,12 +1687,12 @@ class Multisig_Wallet(Deterministic_Wallet):
     def get_keystores(self):
         return [self.keystores[i] for i in sorted(self.keystores.keys())]
 
-    def update_password(self, old_pw, new_pw):
+    def update_password(self, old_pw, new_pw, encrypt=False):
         for name, keystore in self.keystores.items():
             if keystore.can_change_password():
                 keystore.update_password(old_pw, new_pw)
                 self.storage.put(name, keystore.dump())
-        self.storage.put('use_encryption', (new_pw is not None))
+        self.storage.set_password(new_pw, encrypt)
 
     def check_password(self, password):
         self.keystore.check_password(password)
@@ -1678,15 +1716,11 @@ class Multisig_Wallet(Deterministic_Wallet):
         return ''.join(sorted(self.get_master_public_keys()))
 
     def add_input_sig_info(self, txin, address):
-        txin['derivation'] = derivation = self.get_address_index(address)
-        pubkeys = self.get_pubkeys(*derivation)
-        x_pubkeys = [k.get_xpubkey(*derivation) for k in self.get_keystores()]
-        # sort pubkeys and x_pubkeys, using the order of pubkeys
-        pubkeys, x_pubkeys = zip(*sorted(zip(pubkeys, x_pubkeys)))
-        txin['pubkeys'] = list(pubkeys)
-        txin['x_pubkeys'] = list(x_pubkeys)
-        txin['signatures'] = [None] * len(pubkeys)
-        txin['redeemScript'] = self.redeem_script(*derivation)
+        derivation = self.get_address_index(address)
+        # extended pubkeys
+        txin['x_pubkeys'] = [k.get_xpubkey(*derivation) for k in self.get_keystores()]
+        # we need n place holders
+        txin['signatures'] = [None] * self.n
         txin['num_sig'] = self.m
 
 
