@@ -6,12 +6,13 @@ import sys
 import traceback
 
 import electrum
-from electrum.bitcoin import EncodeBase58Check, DecodeBase58Check, TYPE_ADDRESS, int_to_hex, var_int
+from electrum import bitcoin
+from electrum.bitcoin import TYPE_ADDRESS, int_to_hex, var_int
 from electrum.i18n import _
 from electrum.plugins import BasePlugin, hook
 from electrum.keystore import Hardware_KeyStore, parse_xpubkey
 from ..hw_wallet import HW_PluginBase
-from electrum.util import format_satoshis_plain, print_error
+from electrum.util import format_satoshis_plain, print_error, is_verbose
 
 try:
     import hid
@@ -22,7 +23,7 @@ try:
     from btchip.btchipFirmwareWizard import checkFirmware, updateFirmware
     from btchip.btchipException import BTChipException
     BTCHIP = True
-    BTCHIP_DEBUG = False
+    BTCHIP_DEBUG = is_verbose
 except ImportError:
     BTCHIP = False
 
@@ -78,15 +79,11 @@ class Ledger_Client():
                 childnum = int(lastChild[0])
             else:
                 childnum = 0x80000000 | int(lastChild[0])
-            xpub = "0488B21E".decode('hex') + chr(depth) + self.i4b(fingerprint) + self.i4b(childnum) + str(nodeData['chainCode']) + str(publicKey)
-        except Exception, e:
-            #self.give_error(e, True)
+            xpub = bitcoin.serialize_xpub(0, str(nodeData['chainCode']), str(publicKey), depth, self.i4b(fingerprint), self.i4b(childnum))
+            return xpub
+        except Exception as e:
+            print_error(e)
             return None
-        finally:
-            #self.handler.clear_dialog()
-            pass
-
-        return EncodeBase58Check(xpub)
 
     def has_detached_pin_support(self, client):
         try:
@@ -108,9 +105,13 @@ class Ledger_Client():
                 return True
             raise e
 
+    def supports_multi_output(self):
+        return self.multiOutputSupported
+
     def perform_hw1_preflight(self):
         try:
             firmware = self.dongleObject.getFirmwareVersion()['version'].split(".")
+            self.multiOutputSupported = int(firmware[0]) >= 1 and int(firmware[1]) >= 1 and int(firmware[2]) >= 4
             if not checkFirmware(firmware):
                 self.dongleObject.dongle.close()
                 raise Exception("HW1 firmware version too old. Please update at https://www.ledgerwallet.com")
@@ -180,6 +181,9 @@ class Ledger_KeyStore(Hardware_KeyStore):
         return self.derivation        
 
     def get_client(self):
+        return self.plugin.get_client(self).dongleObject
+    
+    def get_client_electrum(self):
         return self.plugin.get_client(self)
     
     def give_error(self, message, clear_client = False):
@@ -264,7 +268,7 @@ class Ledger_KeyStore(Hardware_KeyStore):
         output = None
         outputAmount = None
         p2shTransaction = False
-	reorganize = False
+    	reorganize = False
         pin = ""
         self.get_client() # prompt for the PIN before displaying the dialog if necessary
 
@@ -274,10 +278,11 @@ class Ledger_KeyStore(Hardware_KeyStore):
             if txin.get('is_coinbase'):
                 self.give_error("Coinbase not supported")     # should never happen
 
-            if len(txin['pubkeys']) > 1:
+            if txin['type'] in ['p2sh']:
                 p2shTransaction = True
 
-            for i, x_pubkey in enumerate(txin['x_pubkeys']):
+            pubkeys, x_pubkeys = tx.get_sorted_pubkeys(txin)
+            for i, x_pubkey in enumerate(x_pubkeys):
                 if x_pubkey in derivations:
                     signingPos = i
                     s = derivations.get(x_pubkey)
@@ -288,12 +293,12 @@ class Ledger_KeyStore(Hardware_KeyStore):
 
             inputs.append([txin['prev_tx'].raw, txin['prevout_n'], txin.get('redeemScript'), txin['prevout_hash'], signingPos ])
             inputsPaths.append(hwAddress)
-            pubKeys.append(txin['pubkeys'])
+            pubKeys.append(pubkeys)
 
         # Sanity check
         if p2shTransaction:
-            for txinput in tx.inputs():
-                if len(txinput['pubkeys']) < 2:
+            for txin in tx.inputs():
+                if txin['type'] != 'p2sh':
                     self.give_error("P2SH / regular input mixed in same transaction not supported") # should never happen
 
         txOutput = var_int(len(tx.outputs()))
@@ -307,8 +312,9 @@ class Ledger_KeyStore(Hardware_KeyStore):
 
         # Recognize outputs - only one output and one change is authorized
         if not p2shTransaction:
-            if len(tx.outputs()) > 2: # should never happen
-                self.give_error("Transaction with more than 2 outputs not supported")
+            if not self.get_client_electrum().supports_multi_output():
+                if len(tx.outputs()) > 2:
+                    self.give_error("Transaction with more than 2 outputs not supported")
             for _type, address, amount in tx.outputs():
                 assert _type == TYPE_ADDRESS
                 info = tx.output_info.get(address)
@@ -470,6 +476,5 @@ class LedgerPlugin(HW_PluginBase):
         #if client:
         #    client.used()
         if client <> None:
-            client.checkDevice()
-            client = client.dongleObject            
+            client.checkDevice()                    
         return client
